"""
The Engine is the entry-point for autocron.

To use the Engine create an instance and call ``.start(filename)`` on
this instance with the database filename to use as argument. The
function ``start(filename)`` in the ``__init__.py`` module of autocron
handles this.
"""

import pathlib
import signal
import subprocess
import sys
import threading
import time
from types import SimpleNamespace

from .sql_interface import SQLiteInterface


WORKER_MODULE_NAME = "worker.py"
WORKER_START_DELAY = 0.2


def start_subprocess(database_file):
    """
    Starts the worker process in a detached subprocess. The
    `database_file` is a string with an absolute or relative path to the
    database in use.
    """
    worker_file = pathlib.Path(__file__).parent / WORKER_MODULE_NAME
    cmd = [sys.executable, worker_file]
    if database_file:
        cmd.append(database_file)
    cwd = pathlib.Path.cwd()
    return subprocess.Popen(cmd, cwd=cwd)


def run_worker_monitor(exit_event, database_file):
    """
    Starts the worker processes and monitors that the workers are up and
    running. Restart workers if necessary. This function must run in a
    separate thread. The 'exit_event' is a threading.Event() instance
    and the `database_file` is a string with an absolute or relative
    path to the database in use. If the monitor receives an exit_event
    the running workers are terminated and the function will return,
    terminating its own thread as well.
    """
    processes = []
    interface = SQLiteInterface()
    interface.db_name = database_file
    max_workers = interface.get_max_workers()
    for _ in range(max_workers):
        process = start_subprocess(database_file)
        processes.append(SimpleNamespace(pid=process.pid, process=process))
        # don't start multiple workers too fast one after the other
        # because they may run into a sqlite write-lock. This will not cause
        # a wrong behaviour on starting, monitoring and stopping the workers
        # but can lead to a wrong statistic in the autocron admin/setting.
        time.sleep(WORKER_START_DELAY)
    while True:
        for entry in processes:
            if entry.process.poll() is not None:
                # trouble: process is not running any more.
                # deregister the terminated process from the setting
                # and start a new process.
                interface.decrement_running_workers(entry.pid)
                new_process = start_subprocess(database_file)
                entry.pid = new_process.pid
                entry.process = new_process
                # in case more than one process needs a restart:
                time.sleep(WORKER_START_DELAY)
        idle_time = interface.get_monitor_idle_time()
        if exit_event.wait(timeout=idle_time):
            break
    for entry in processes:
        entry.process.terminate()


class Engine:
    """
    The Engine is the entry-point for autocron. Starting the engine will
    start the monitor thread. The monitor-thread in turn starts and
    supervise the workers. (The interface argument is used for testing.)
    """
    def __init__(self, interface=None):
        # the ìnterface argument is for testing. In production this
        # argument is not provided for initialization.
        self.interface = interface if interface else SQLiteInterface()
        self.exit_event = None
        self.monitor_thread = None
        self.orig_signal_handlers = {}

    def set_signal_handlers(self):
        """
        Set self._terminate() as handler for a couple of
        termination-signals and store the orinal handlers for this
        signals.
        """
        signalnums = [
            signal.SIGINT,
            signal.SIGTERM,
            signal.SIGABRT,
            signal.SIGHUP,  # availablity: Unix
        ]
        for signalnum in signalnums:
            self.orig_signal_handlers[signalnum] = signal.getsignal(signalnum)
            signal.signal(signalnum, self._terminate)

    def start(self, database_file):
        """
        Starts the autocron worker in case autocron is active and no
        other application process has already started the workers. The
        *database_file* argument is a string with the name of the
        database to use (like "the_application.db") or a Path instance.
        If the name represents a relative path, the database is stored
        in the ``~/.autocron`` directory. This directory will get
        created if not already existing. If the name represents an
        absolute path, then this path will be used as is. In this case
        all directories of the path must exist.

        The function returns a boolean: True if workers have been
        started and False otherwise. A return value of False does not
        mean, that no workers are running – another application process
        may have be first.
        """
        self.interface.init_database(database_file)
        if (
            self.interface.autocron_lock_is_set
            or self.interface.monitor_lock_flag_is_set
        ):
            # in both cases start is not allowed
            return False
        self.interface.set_monitor_lock_flag(True)
        # this is a safety check for not starting more than
        # one monitor thread:
        if not self.monitor_thread:
            self.set_signal_handlers()
            self.exit_event = threading.Event()
            self.monitor_thread = threading.Thread(
                target=run_worker_monitor,
                args=(self.exit_event, database_file)
            )
            self.monitor_thread.start()
        return True

    def stop(self):
        """
        Shut down the monitor-thread which in turn will stop all running
        workers. Also release the monitor_lock flag.
        """
        if self.monitor_thread:
            # check for self.exit_event for a test-scenario.
            # in production if self.monitor_thread is not None
            # self.exit_event is also not None
            if self.exit_event:
                self.exit_event.set()
            self.monitor_thread = None
            settings = self.interface.get_settings()
            settings.monitor_lock = False
            # in case of a crash the workers will terminate but
            # may not reliable reset their states. So do it here:
            settings.running_workers = 0
            settings.worker_pids = ""
            self.interface.set_settings(settings)
            self.interface.delete_cronjobs()

    def _terminate(self, signalnum, stackframe=None):
        """
        Terminate autocron by calling the engine.stop() method.
        Afterward reraise the signal again for the original
        signal-handler.
        """
        # stackframe may be given to the signal-handler, but is unused
        # pylint: disable=unused-argument
        self.stop()
        signal.signal(signalnum, self.orig_signal_handlers[signalnum])
        signal.raise_signal(signalnum)  # requires Python >= 3.8



"""
schedule.py

for cronjobs: from the last schedule (or now) and a given crontab,
calculates the next schedule.
"""
# pylint: disable=too-many-arguments

import calendar
import datetime
from datetime import timedelta
import re


class Sequencer:
    """
    Data-descriptor for runtime type-checking and -conversion.
    """
    def __init__(self):
        self.name = None
        self.storage = None

    def __set_name__(self, cls, name):
        self.name = name
        self.storage = f"_{name}"

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.storage)

    def __set__(self, obj, value):
        if value is None:
            pass
        elif isinstance(value, int):
            value = [value]
        elif not isinstance(value, (tuple, list)):
            msg = f"Argument '{self.name}' must be of type int, tuple or list, "\
                  f"got {type(value)} instead."
            raise TypeError(msg)
        setattr(obj, self.storage, value)


class CronScheduler:
    """
    Schedules a cron task.

    Usage:

    cs = CronScheduler()
    next_schedule = cs.get_next_schedule()
    """

    minutes = Sequencer()
    hours = Sequencer()
    dow = Sequencer()
    month = Sequencer()
    dom = Sequencer()

    def __init__(self, last_schedule=None, minutes=None, hours=None,
                 dow=None, months=None, dom=None, crontab=None):
        """
        Inits the scheduler with values according the crontab-format.

        All arguments are optional. If no arguments are given the task
        is scheduled from now on every minute.

        last_schedule: last time the according tasks has run
        minutes: list of integers when a task should run in an hour [0-59]
        hours: list of integers when a task should run during a day [0-23]
        dow: list of integers for the days of week when a task should
        run [0-6] (0 for monday up to 6 for sunday).
        month: list of integers for the months a task should run [1-12]
        dom: list of intergs for the day in a month a task should run [1-31]
        """
        self.last_schedule = last_schedule or datetime.datetime.now()
        self.minutes = minutes
        self.hours = hours
        self.dow = dow
        self.months = months
        self.dom = dom
        self.parse_crontab(crontab)

    def parse_crontab(self, crontab):
        """
        Parses a simple crontab-string with five patterns:

        Minute [0,59]
        Hour [0,23]
        Day of the month [1,31]
        Month of the year [1,12]
        Day of the week ([0,6] with 0=Monday)

        Each of these patterns can be either an asterisk (meaning all
        valid values), an element, or a list of elements separated by
        commas. An element shall be either a number or two numbers
        separated by a hyphen (meaning an inclusive range).

        A day-range for a month like 1-4,12,20-25 is also ok. But keep
        in mind that no spaces are allowed as spaces are the separators
        between the patterns.

        Some examples:

        * * * * *               runs every minute
        5 * * * *               runs every five minutes
        30 7 0 4,7 10-15        runs at 7:30 on mondays and also from the
                                10th to 15th of a month, but only in april
                                and july

        May raise errors if the crontab could not be parsed. The
        crontab-content is *not* checked for correctness).
        """
        def crontab_values():
            for item in crontab.split():
                item = item.strip()
                if item == '*':
                    yield None
                    continue
                inner = []
                for element in item.split(','):
                    match_object = re.match(r'(\d+)-(\d+)', element)
                    if match_object:
                        inner.extend(list(range(
                            int(match_object.group(1)),
                            int(match_object.group(2))+1
                        )))
                    else:
                        inner.append(int(element))
                yield inner

        if crontab:
            crontab_value = crontab_values()
            # some magic: dynamic setting of instance attributes
            for name in ("minutes", "hours", "dow", "months", "dom"):
                setattr(self, name, next(crontab_value))

    def get_next_schedule(self, last_schedule=None):
        """
        Returns the next schedule based on the current date as a
        datetime-object.
        """
        last_schedule = last_schedule or self.last_schedule
        next_schedule = self.find_next_schedule(last_schedule)
        if not self.month_allowed(next_schedule):
            schedule = self.set_allowed_month(next_schedule)
            return self.get_next_schedule(schedule)
        return next_schedule

    def set_allowed_month(self, schedule):
        """
        Modifies the schedule to the first day of the next allowed
        month and returns this new schedule.
        """
        month = get_next_value(schedule.month, self.months)
        if month > schedule.month:
            year = schedule.year
        else:
            year = schedule.year + 1
        return datetime.datetime(year, month, 1)

    def month_allowed(self, schedule):
        """
        Returns a boolean whether the scheduled month is allowed by the
        crontab data.
        """
        if not self.months:
            return True
        return schedule.month in self.months

    def day_allowed(self, schedule):
        """
        Returns a boolean whether the scheduled day is allowed by the
        crontab data.
        """
        if not self.dom and not self.dow:
            # every day is allowed
            return True
        if self.dom and schedule.day in self.dom:
            # day of month is allowed
            return True
        weekday = calendar.weekday(schedule.year, schedule.month, schedule.day)
        if self.dow and weekday in self.dow:
            # weekday is allowed
            return True
        return False

    def hour_allowed(self, schedule):
        """
        Returns a boolean whether the scheduled hour is allowed by the
        crontab data.
        """
        if not self.hours:
            return True
        return schedule.hour in self.hours

    def find_next_schedule(self, last_schedule):
        """
        Finds the next allowed day and time according to the crontab
        data. It has to be assumed that last_schedule may not be a valid
        schedule (may be initialized with now()).
        Returns a datetime-object.
        """
        # short cut
        ls = last_schedule  # pylint: disable=invalid-name
        next_minute = self.get_next_minute(ls)
        if next_minute > ls.minute:
            # same day, same hour
            delta = next_minute - ls.minute
            if self.hour_allowed(ls):
                return ls + timedelta(minutes=delta)
        next_hour = self.get_next_hour(ls)
        if next_hour > ls.hour:
            # same day
            if self.day_allowed(ls):
                delta_hours = next_hour - ls.hour
                delta_minutes = next_minute - ls.minute
                return ls + timedelta(hours=delta_hours, minutes=delta_minutes)
        if not self.dow and not self.dom:
            # next day
            schedule = self.get_next_day(ls)
        elif self.dom and not self.dow:
            # find next allowed day of month
            schedule = self.get_next_dom(ls)
        elif not self.dom and self.dow:
            # find next allowed weekday
            schedule = self.get_next_dow(ls)
        elif self.dom and self.dow:
            # find first next allowed day of month
            schedule_dom = self.get_next_dom(ls)
            schedule_dow = self.get_next_dow(ls)
            schedule = min(schedule_dom, schedule_dow)
        return schedule + timedelta(
            hours=next_hour-ls.hour,
            minutes=next_minute-ls.minute)

    def get_next_minute(self, last_schedule):
        """
        Returns the next minute a task should run according to
        last_schedule (a datetime object) and the content of
        self.minutes.
        """
        if self.minutes:
            minute = get_next_value(last_schedule.minute, self.minutes)
        else:
            minute = last_schedule.minute + 1
            if minute > 59:
                minute = 0
        return minute

    def get_next_hour(self, last_schedule):
        """
        Returns the next hour a task should run according to
        last_schedule ( a datetime object) and the content of
        self.hours.
        """
        if self.hours:
            hour = get_next_value(last_schedule.hour, self.hours)
        else:
            hour = last_schedule.hour + 1
            if hour > 23:
                hour = 0
        return hour

    @staticmethod
    def get_next_day(last_schedule):
        """
        Increments the day according to last_schedule (a
        datetime-object) and returns the new date as datetime-object.
        """
        return last_schedule + datetime.timedelta(days=1)

    def get_next_dom(self, last_schedule):
        """
        Calculates the next allowed day and returns the result as a
        datetime-object.
        """
        # method gets not called if self.dom is None or empty
        next_day = get_next_value(last_schedule.day, self.dom)
        day = last_schedule.day
        if next_day > day:
            delta = next_day - day
        else:
            _, max_days = calendar.monthrange(
                last_schedule.year, last_schedule.month)
            delta = max_days - day + next_day
        return last_schedule + datetime.timedelta(days=delta)

    def get_next_dow(self, last_schedule):
        """
        Calculates the next allowed weekday and returns the result as a
        datetime-object.
        """
        weekday = calendar.weekday(
            last_schedule.year, last_schedule.month, last_schedule.day)
        next_weekday = get_next_value(weekday, self.dow)
        if next_weekday > weekday:
            delta = next_weekday - weekday
        else:
            delta = 7 - weekday + next_weekday
        return last_schedule + datetime.timedelta(days=delta)


def get_next_value(value, values):
    """
    Returns the next value from values which is larger then value or the
    first item from the sequence values.
    """
    for item in values:
        if item > value:
            return item
    return values[0]



"""
admin.py

Administration tool to access the database.
"""

import argparse

from autocron.sql_interface import (
    SQLiteInterface,
    DEFAULT_MAX_WORKERS,
    DEFAULT_RUNNING_WORKERS,
    DEFAULT_MONITOR_LOCK,
    DEFAULT_AUTOCRON_LOCK,
)

interface = SQLiteInterface()


PGM_NAME = "autocron command line tool"
PGM_DESCRIPTION = """
    Allows access to the autocron database to change default
    settings and to inspect waiting tasks and stored results.
"""


def _format_item_list(data, separator="\n", divider=":", column_width=None):
    """
    Helperfunction: takes a string representing a list of data-pairs
    separated by a given separator, with the data in the data-pair
    separated by the given divider (i.e. "1:2\n3:4" etc). Formats the
    data pairs according to the longest first item, so that the dividers
    (the colons in the exaple data given above) are vertical aligned.
    Returns the new formatted string.
    """
    keys = []
    values = []
    for line in data.split(separator):
        key, value = line.split(divider, 1)
        keys.append(key)
        values.append(value)
    if column_width is None:
        # here is pylint wrong:
        # pylint: disable=unnecessary-lambda
        column_width = len(max(keys, key=lambda x: len(x)))
    template = f"{{:{column_width}}}{divider} {{}}"
    return separator.join(template.format(k, v) for k, v in zip(keys, values))


def report_info():
    """Report the settings and statistics about tasks and results."""
    settings = interface.get_settings()
    task_num = interface.count_tasks()
    result_num = interface.count_results()
    report = f"{settings}\ntasks:{task_num}\nresults:{result_num}"
    formated_report = _format_item_list(report)
    output = f"\nSettings\n{formated_report}\n"
    print(output)


def _report_tasks(tasks, task_type="tasks"):
    """
    Helperfunction: takes a list of Hybridnamespace intstances and print
    them nicely formated to stdout.
    """
    print(f"\n{task_type} found: {len(tasks)}")
    divider = "-" * 50
    for task in tasks:
        task = _format_item_list(str(task))
        print(f"{divider}\n{task}")
    print(divider)
    print()


def report_tasks():
    """Report all stored tasks."""
    _report_tasks(interface.get_tasks())


def report_tasks_on_due():
    """Report all tasks waiting for execution and are on due."""
    _report_tasks(interface.get_tasks_on_due())


def report_cron_tasks():
    """Report all task which are cron-jobs."""
    _report_tasks(interface.get_crontasks())


def report_results():
    """Report all available results."""
    results = interface.get_results()
    _report_tasks(results, task_type="results")


def print_usage():
    """Print program-description and hint how to get help."""
    print(f"\n{PGM_NAME}\n{PGM_DESCRIPTION}\nuse option -h for help.\n\n")


def reset_defaults():
    """Reset the settings with the default values."""
    settings = interface.get_settings()
    settings.max_workers = DEFAULT_MAX_WORKERS
    settings.running_workers = DEFAULT_RUNNING_WORKERS
    settings.monitor_lock = DEFAULT_MONITOR_LOCK
    settings.autocron_lock = DEFAULT_AUTOCRON_LOCK
    interface.set_settings(settings)
    print("\nReset settings default values:")
    report_info()


def set_max_workers(workers):
    """Change value for the number of allowed max_workers."""
    settings = interface.get_settings()
    settings.max_workers = workers
    interface.set_settings(settings)
    print(f"Set max_workers to {workers}")


def set_worker_idle_time(idle_time):
    """
    Set the idle time of the worker in seconds. This is the time the
    worker sleeps when no new tasks are on due.
    """
    settings = interface.get_settings()
    settings.worker_idle_time = idle_time
    interface.set_settings(settings)
    print(f"Set worker idle time to {idle_time} seconds")


def set_monitor_idle_time(idle_time):
    """
    Set the idle time of the worker in seconds. This is the time the
    worker sleeps when no new tasks are on due.
    """
    settings = interface.get_settings()
    settings.monitor_idle_time = idle_time
    interface.set_settings(settings)
    print(f"Set monitor idle time to {idle_time} seconds")


def set_result_ttl(ttl):
    """
    Set the result time to life in seconds. ttl is an integer. This is
    the timespan a result will get stored in the database.
    """
    settings = interface.get_settings()
    settings.result_ttl = ttl
    interface.set_settings(settings)
    print(f"Set result-ttl to {ttl} seconds")


def _convert_flag(flag):
    """
    flag can be a string with 'true'', 'false' or 'on'', 'off''.
    Returns a boolean.
    """
    flag = flag.lower()
    return flag in {"true", "on"}


def set_monitor_lock(flag):
    """Set monitor_lock flag."""
    flag = _convert_flag(flag)
    interface.set_monitor_lock_flag(flag)
    print(f"Set monitor lock to {flag}")


def set_autocron_lock(flag):
    """Set autocron_lock flag."""
    flag = _convert_flag(flag)
    settings = interface.get_settings()
    settings.autocron_lock = flag
    interface.set_settings(settings)
    print(f"Set autocron lock to {flag}")


def delete_database():
    """
    Delete the database and create the database again with the default
    settings.
    """
    answer = input("Sure to delete the current database? [y/n]: ")
    if answer.lower() == 'y':
        interface.db_name.unlink()
        # this could be all, because at the next autocron import
        # a new database will get created.
        # However, create a new database here and now:
        # pylint: disable=protected-access
        interface._init_database()
    else:
        print("abort command")


def get_command_line_arguments():
    """Get the command line arguments."""
    parser = argparse.ArgumentParser(
        prog=PGM_NAME,
        description=PGM_DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "database",
        help="database name or path."
    )
    parser.add_argument(
        "-i", "--info",
        action="store_true",
        help="provide information about the settings, number of waiting tasks "
             "and result entries."
    )
    parser.add_argument(
        "--reset-defaults",
        dest="reset_defaults",
        action="store_true",
        help="restore the default settings: max_workers=1, running_workers=0, "
             "monitor_lock=0, autocron_lock=0."
    )
    parser.add_argument(
        "--delete-database",
        dest="delete_database",
        action="store_true",
        help="delete the current database and create a new clean one with "
             "the default settings."
    )
    parser.add_argument(
        "--set-max-workers",
        dest="max_workers",
        type=int,
        help="set number of maximum worker processes."
    )
    parser.add_argument(
        "--set-worker-idle-time",
        dest="worker_idle_time",
        type=float,
        help="set worker idle time in seconds."
    )
    parser.add_argument(
        "--set-monitor-idle-time",
        dest="monitor_idle_time",
        type=float,
        help="set monitor idle time in seconds."
    )
    parser.add_argument(
        "--set-monitor-lock",
        dest="monitor_lock",
        help="set monitor lock flag: [true|false or on|off]."
    )
    parser.add_argument(
        "--set-autocron-lock",
        dest="autocron_lock",
        help="set autocron lock flag: [true|false or on|off]."
    )
    parser.add_argument(
        "--set-result-ttl",
        dest="result_ttl",
        type=int,
        help="set result time to life in seconds."
    )
    parser.add_argument(
        "-t", "--get-tasks",
        dest="get_tasks",
        action="store_true",
        help="list all waiting tasks."
    )
    parser.add_argument(
        "-d", "--get-tasks-on-due",
        dest="get_tasks_on_due",
        action="store_true",
        help="list all waiting tasks that are on due."
    )
    parser.add_argument(
        "-c", "--get-cron-tasks",
        dest="get_cron_tasks",
        action="store_true",
        help="list all task which are cronjobs."
    )
    parser.add_argument(
        "-r", "--get-results",
        dest="get_results",
        action="store_true",
        help="list all available results."
    )
    return parser.parse_args()


def main(args=None):
    """entry point."""
    # pylint: disable=too-many-branches
    if not args:
        args = get_command_line_arguments()
    # like the worker just use the database "as is",
    # by not running the initialization.
    interface.db_name = args.database
    if args.info:
        report_info()
    elif args.reset_defaults:
        reset_defaults()
    elif args.max_workers:
        set_max_workers(args.max_workers)
    elif args.get_tasks:
        report_tasks()
    elif args.get_tasks_on_due:
        report_tasks_on_due()
    elif args.get_cron_tasks:
        report_cron_tasks()
    elif args.get_results:
        report_results()
    elif args.delete_database:
        delete_database()
    elif args.monitor_lock:
        set_monitor_lock(args.monitor_lock)
    elif args.autocron_lock:
        set_autocron_lock(args.autocron_lock)
    elif args.worker_idle_time:
        set_worker_idle_time(args.worker_idle_time)
    elif args.monitor_idle_time:
        set_worker_idle_time(args.monitor_idle_time)
    elif args.result_ttl:
        set_result_ttl(args.result_ttl)
    else:
        print_usage()


if __name__ == "__main__":
    main(get_command_line_arguments())



"""
autocron:

simple asynchronous background task handling with no dependencies beside
the standard-library.
"""
from .decorators import (
    cron,
    delay,
)
from .engine import Engine
from .sql_interface import SQLiteInterface


__all__ = ["cron", "delay", "start", "stop", "get_results"]
__version__ = "0.9.1"

_engine = Engine()
_interface = SQLiteInterface()


def start(database_file):
    """
    Call this from the framework of choice to explicitly activate
    autocron. ``database_file`` is a string with the file-name of the
    database. The file gets stored in the ``~.autocron/`` directory. If
    this directory does not exist, it will get created. The file-name can
    also be an absolute path so the file will get stored elsewere. In
    this case all directories in the path must exist.
    """
    _engine.start(database_file=database_file)


def stop():
    """
    Stops autocron explicitly. On receiving a termination-signal
    autocron invokes a shutdown sequence to stop the workers, so calling
    ``stop`` is normalwise not required.
    """
    _engine.stop()


def get_results():
    """
    Returns a list with all result entries from delayed functions.
    """
    return _interface.get_results()



"""
Implementatiom of the decorators ``cron`` and ``delay`` for running
recurring task and to delegate long running tasks to a background
process.
"""

import uuid

from .schedule import CronScheduler
from .sql_interface import SQLiteInterface, TaskResult


# default: run every minute:
DEFAULT_CRONTAB = "* * * * *"

interface = SQLiteInterface()


# pylint: disable=too-many-arguments
def cron(crontab=None,
         minutes=None,
         hours=None,
         dow=None,
         months=None,
         dom=None):
    """
    Decorator for a cronjob. Functions running cronjobs should not get
    called from the main program and therefore don't get arguments.
    Example usage for a cronjob to run every hour, at the beginning of
    the hour:

    >>> @cron("0 * * * *")
    >>> def some_callable():
    >>>    # do periodic stuff here ...

    The decorator can take a couple of arguments but if just the first
    argument `crontab` is given then all other arguments are ignored. To
    use the other arguments instead, provide them all as
    keyword-arguments. If no arguments are given the default-crontab
    ``(* * * * *)`` is used to execute a task every minute.

    :crontab:
        a string representing a valid crontab. See:
        `https://en.wikipedia.org/wiki/Cron#CRON_expression
        <https://en.wikipedia.org/wiki/Cron#CRON_expression>`_ with the
        restriction that only integers and the special signs (* , -) are
        allowed. Some examples ::

            The order of arguments is:
            'minutes hours dow months dom'

            '* * * * *':       runs every minute
                               (same as @periodic_task(seconds=60))
            '15,30 7 * * *':   runs every day at 7:15 and 7:30
            '* 9 0 4,7 10-15': runs at 9:00 every monday and
                               from the 10th to the 15th of a month
                               but only in April and July.


    :minutes:
        list of minutes during an hour when the task should run. Valid
        entries are integers in the range 0-59. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every minute.

    :hours:
        list of hours during a day when the task should run. Valid
        entries are integers in the range 0-23. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every hour.

    :dow:
        days of week. A list of integers from 0 to 6 with Monday as 0.
        The task runs only on the given weekdays. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every day of the week.

    :months:
        list of month during a year when the task should run. Valid
        entries are integers in the range 1-12. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every month.

    :dom:
        list of days in an month the task should run. Valid entries are
        integers in the range 1-31. Defaults to None which is the same
        as ``*`` in a crontab, meaning that the task gets executed every
        day.

    If neither *dom* nor *dow* are given, then the task will run every
    day of a month. If one of both is set, then the given restrictions
    apply. If both are set, then the allowed days complement each other.
    """
    # set crontab to default if no other arguments are given:
    if not any(locals().values()):
        crontab = DEFAULT_CRONTAB

    def wrapper(func):
        scheduler = CronScheduler(
            minutes=minutes,
            hours=hours,
            dow=dow,
            months=months,
            dom=dom,
            crontab=crontab
        )
        schedule = scheduler.get_next_schedule()
        interface.register_callable(
            func,
            schedule=schedule,
            crontab=crontab,
            unique=True  # don't register cron-tasks twice
        )
        return func

    return wrapper


def delay(func):
    """
    Decorator for a delayed task. Apply this as:

    >>> @delay
    >>> def sendmail(recipient, message):
    >>>     # conde goes here ...

    The decorator does not take any arguments. Calling ``sendmail()``
    will return from the call immediately and this callable will get
    executed later in another process.
    """

    def wrapper(*args, **kwargs):
        if interface.accept_registrations:
            # active: return TaskResult in waiting state
            uid = uuid.uuid4().hex
            data = {"args": args, "kwargs": kwargs, "uuid": uid}
            interface.register_callable(func, **data)
            result = TaskResult.from_registration(uid, interface)
        elif interface.autocron_lock_is_set:
            # inactive: return TaskResult in ready state
            # by calling the wrapped function. This will keep the
            # type of the returned value consistant for the application.
            result = TaskResult.from_function_call(func, *args, **kwargs)
        else:
            # active but registration not allowed.
            # this is the wrapper as seen by a worker process:
            # return result from the original callable
            result = func(*args, **kwargs)
        return result

    return wrapper



"""
worker.py

worker class for handling cron and delegated tasks.
"""

import importlib
import os
import signal
import sys
import time

from autocron.schedule import CronScheduler
from autocron import sql_interface


WORKER_IDLE_TIME = 4.0  # seconds


class Worker:
    """
    Runs in a separate process for task-handling.
    Gets supervised and monitored from the engine.
    """
    def __init__(self, database_filename):
        self.active = True
        self.result = None
        self.error_message = None
        signal.signal(signal.SIGINT, self.terminate)
        signal.signal(signal.SIGTERM, self.terminate)
        # Get a SQLiteInterface instance 'as is' without the
        # initialization step to clean up tasks.
        # Providing the databasename will set the dtatabase
        # to initialized-state. So no new tasks are registered
        # from a worker.
        # (keep in mind that every worker runs in a separate process.)
        self.interface = sql_interface.SQLiteInterface()
        self.interface.db_name=database_filename
        self.interface.accept_registrations = False

    def terminate(self, *args):  # pylint: disable=unused-argument
        """
        Signal handler to stop the process, terminates the loop in
        `run()`.
        """
        self.active = False

    def run(self):
        """
        Main event loop for the worker. Takes callables and processes
        them as long as callables are available. Otherwise keep idle.
        """
        pid = os.getpid()
        self.interface.increment_running_workers(pid)
        while self.active:
            if not self.handle_tasks():
                # nothing to do, check for results to delete:
                self.interface.delete_outdated_results()
                time.sleep(self.interface.get_worker_idle_time())
        self.interface.decrement_running_workers(pid)

    def handle_tasks(self):
        """
        Checks for tasks and process them. If there are no tasks to
        handle the method return `False` indicating that the main loop
        can switch to idle state. If  tasks have been handled, the
        method return `True` to indicate that meanwhile more tasks may be
        waiting.
        """
        tasks = self.interface.get_tasks_on_due(
            status=sql_interface.TASK_STATUS_WAITING,
            new_status=sql_interface.TASK_STATUS_PROCESSING
        )
        if tasks:
            for task in tasks:
                if self.active is False:
                    # terminate as soon as possible
                    return True
                self.error_message = None
                self.result = None
                self.process_task(task)
                self.postprocess_task(task)
            return True
        return False

    def process_task(self, task):
        """
        Handle the given task. The task is a dictionary as returned from
        SQLInterface._fetch_all_callable_entries(cursor):

            {
                "rowid": integer,
                "uuid": string,
                "schedule": datetime,
                "crontab": string,
                "function_module": string,
                "function_name": string,
                "args": tuple(of original datatypes),
                "kwargs": dict(of original datatypes),
            }

        """
        module = importlib.import_module(task.function_module)
        function = getattr(module, task.function_name)
        try:
            self.result = function(*task.args, **task.kwargs)
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.error_message = repr(err)

    def postprocess_task(self, task):
        """
        Delete or update the task and do something with the result or
        error-message.
        """
        if task.uuid:
            # if the task has a uuid, store the result / error-message
            self.interface.update_result(
                uuid=task.uuid,
                result=self.result,
                error_message=self.error_message
            )
        if task.crontab:
            # if the task has a crontab calculate new schedule
            # and update the task-entry
            scheduler = CronScheduler(crontab=task.crontab)
            schedule = scheduler.get_next_schedule()
            self.interface.update_crontask_schedule(
                rowid=task.rowid,
                schedule=schedule
            )
        else:
            # not a cronjob: delete the task from the db
            self.interface.delete_callable(task)


def start_worker():
    """subprocess entry-point"""
    # insert cwd of hosting application to pythonpath
    sys.path.insert(0, os.getcwd())
    # engine provides the database name as first argument:
    database_filename = sys.argv[1]
    worker = Worker(database_filename)
    worker.run()


if __name__ == "__main__":
    start_worker()



"""
test_sql_interface.py

testcode for sql-actions and for decorators
"""

import collections
import datetime
import pathlib
import sqlite3
import time
import unittest
import uuid

from autocron import decorators
from autocron import sql_interface
from autocron import worker

from autocron.sql_interface import TaskResult


TEST_DB_NAME = "test.db"
ANOTHER_FILE_NAME = "another_file_name.db"


# all labels starting with 'tst_' are test-functions:

def tst_callable(*args, **kwargs):
    return args, kwargs

def tst_add(a, b):
    return a + b

def tst_multiply(a, b):
    return a * b

def tst_cron():
    pass

def tst_delay():
    return 42


class TestSQLInterface(unittest.TestCase):

    def setUp(self):
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(db_name=TEST_DB_NAME)
        self._result_ttl = self.interface._result_ttl

    def tearDown(self):
        pathlib.Path(self.interface.db_name).unlink(missing_ok=True)
        self.interface._result_ttl = self._result_ttl

    def test_storage_location(self):
        path = pathlib.Path.home() / sql_interface.DEFAULT_STORAGE / TEST_DB_NAME
        assert self.interface.db_name == path
        # don't allow setting the db a second time:
        self.interface.init_database(db_name=ANOTHER_FILE_NAME)
        assert self.interface.db_name == path

    def test_storage_location_absolute(self):
        pathlib.Path(self.interface.db_name).unlink()
        self.interface.db_name = None
        path = pathlib.Path.cwd() / ANOTHER_FILE_NAME
        self.interface.init_database(db_name=path)
        assert self.interface.db_name == path

    def test_storage(self):
        entries = self.interface.get_tasks_on_due()
        self.assertFalse(list(entries))
        self.interface.register_callable(tst_callable)
        entries = self.interface.get_tasks_on_due()
        assert len(entries) == 1

    def test_entry_signature(self):
        self.interface.register_callable(tst_callable)
        entries = self.interface.get_tasks_on_due()
        obj = entries[0]
        assert isinstance(obj, sql_interface.HybridNamespace) is True
        assert obj["function_module"] == tst_callable.__module__
        assert obj["function_name"] == tst_callable.__name__

    def test_arguments(self):
        args = ["pi", 3.141]
        kwargs = {"answer": 41, 10: "ten"}
        crontab = "* 1 * * *"
        self.interface.register_callable(
            tst_callable, crontab=crontab, args=args, kwargs=kwargs
        )
        entries = list(self.interface.get_tasks_on_due())
        obj = entries[0]
        assert obj["crontab"] == crontab
        assert obj["args"] == args
        assert obj["kwargs"] == kwargs

    def test_get_tasks(self):
        # test the generic function to select all tasks:
        schedule = datetime.datetime.now() + datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=schedule)
        self.interface.register_callable(tst_callable)
        self.interface.register_callable(tst_multiply, crontab="* * * * *")
        # should return everything:
        entries = self.interface.get_tasks()
        assert len(entries) == 3

    def test_schedules_get_one_of_two(self):
        # register two callables, one with a schedule in the future
        schedule = datetime.datetime.now() + datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=schedule)
        self.interface.register_callable(tst_callable)
        # test to get one callable at due
        entries = self.interface.get_tasks_on_due()
        assert len(entries) == 1

    def test_schedules_get_two_of_two(self):
        # register two callables, both scheduled in the present or past
        schedule = datetime.datetime.now() - datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=schedule)
        self.interface.register_callable(tst_callable)
        # test to get one callable at due
        entries = self.interface.get_tasks_on_due()
        assert len(entries) == 2

    def test_delete(self):
        # register two callables, one with a schedule in the future
        now = datetime.datetime.now()
        future_schedule = now + datetime.timedelta(milliseconds=2)
        past_schedule = now - datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=future_schedule)
        self.interface.register_callable(tst_callable, schedule=past_schedule)
        # test to get the `tst_callable` function on due
        # and delete it from the db
        entry = self.interface.get_tasks_on_due()[0]
        assert entry["function_name"] == tst_callable.__name__
        self.interface.delete_callable(entry)
        # wait and test to get the remaining single entry
        # and check whether it is the `tst_add` function
        time.sleep(0.002)
        entries = self.interface.get_tasks_on_due()
        assert len(entries) == 1
        entry = entries[0]
        assert entry["function_name"] == tst_add.__name__

    def test_get_task_by_signature(self):
        # register two callables, one with a schedule in the future
        schedule = datetime.datetime.now() + datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=schedule)
        self.interface.register_callable(tst_callable)
        # find a nonexistent callable should return an empty generator
        entries = self.interface.get_tasks_by_signature(tst_multiply)
        assert len(entries) == 0
        # find a callable scheduled for the future:
        entries = self.interface.get_tasks_by_signature(tst_add)
        assert len(entries) == 1

    def test_get_tasks_by_signature(self):
        # it is allowed to register the same callables multiple times.
        # regardless of the schedule `get_tasks_by_signature()` should return
        # all entries.
        schedule = datetime.datetime.now() + datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=schedule)
        self.interface.register_callable(tst_add)
        entries = list(self.interface.get_tasks_by_signature(tst_add))
        assert len(entries) == 2

    def test_get_task_on_due_and_set_status(self):
        # set two task which are on due with status WAITING (default).
        # Select them by setting the status to PROCESSED.
        # A second selection should not work.
        schedule = datetime.datetime.now() - datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=schedule)
        self.interface.register_callable(tst_add, schedule=schedule)
        entries = self.interface.get_tasks_on_due(
            status=sql_interface.TASK_STATUS_WAITING,
            new_status=sql_interface.TASK_STATUS_PROCESSING
        )
        assert len(entries) == 2
        # do this a second time should not work
        entries = self.interface.get_tasks_on_due(
            status=sql_interface.TASK_STATUS_WAITING,
            new_status=sql_interface.TASK_STATUS_PROCESSING
        )
        assert len(entries) == 0
        # but the two entries are still there:
        entries = self.interface.get_tasks()
        assert len(entries) == 2

    def test_update_schedule(self):
        # entries like cronjobs should not get deleted from the tasks
        # but updated with the next schedule
        schedule = datetime.datetime.now()
        next_schedule = schedule + datetime.timedelta(seconds=10)
        self.interface.register_callable(tst_add, schedule=schedule)
        entry = self.interface.get_tasks_by_signature(tst_add)[0]
        assert entry["schedule"] == schedule
        self.interface.update_crontask_schedule(entry["rowid"], next_schedule)
        entry = self.interface.get_tasks_by_signature(tst_add)[0]
        assert entry["schedule"] == next_schedule

    def test_update_crontask(self):
        """
        When a crontask is selected for handling because it is 'on due',
        the status changes from WAITING to PROCESSING. After
        task-handling and schedule update the status must get reset to
        WAITING.
        """
        # after adding a crontask the task is in WAITING state:
        self.interface.register_callable(tst_add)
        task = self.interface.get_tasks_by_signature(tst_add)[0]
        assert task.status == sql_interface.TASK_STATUS_WAITING
        # after retrieving on due, the state changes to PROCESSING
        # for the returned task-object and also for the stored task-object:
        task = self.interface.get_tasks_on_due(
            new_status=sql_interface.TASK_STATUS_PROCESSING
        )[0]
        assert task.status == sql_interface.TASK_STATUS_PROCESSING
        task = self.interface.get_tasks_by_signature(tst_add)[0]
        assert task.status == sql_interface.TASK_STATUS_PROCESSING
        # after calling update_crontask_schedule() the status must
        # get reset to WAITING again:
        schedule = datetime.datetime.now()
        rowid = task.rowid
        self.interface.update_crontask_schedule(rowid, schedule)
        task = self.interface.get_tasks_by_signature(tst_add)[0]
        assert task.status == sql_interface.TASK_STATUS_WAITING

    def test_result_by_uuid_no_result(self):
        # result should be None if no entry found
        uuid_ = uuid.uuid4().hex
        result = self.interface.get_result_by_uuid(uuid_)
        assert result is None

    def test_result_by_uuid_result_registered(self):
        uuid_ = uuid.uuid4().hex
        self.interface.register_result(tst_add, uuid=uuid_)
        result = self.interface.get_result_by_uuid(uuid_)
        # return a TaskResult instance:
        assert result is not None
        assert result.is_waiting is True

    def test_update_result_no_error(self):
        answer = 42
        uuid_ = uuid.uuid4().hex
        self.interface.register_result(tst_add, uuid=uuid_)
        self.interface.update_result(uuid_, result=answer)
        result = self.interface.get_result_by_uuid(uuid_)
        assert result.is_ready is True
        assert result.function_result == answer
        # test shortcut for function_result:
        assert result.result == answer

    def test_update_result_with_error(self):
        message = "ValueError: more text here ..."
        uuid_ = uuid.uuid4().hex
        self.interface.register_result(tst_add, uuid=uuid_)
        self.interface.update_result(uuid_, error_message=message)
        result = self.interface.get_result_by_uuid(uuid_)
        assert result.has_error is True

    def test_do_not_delete_waiting_results(self):
        self.interface._result_ttl = datetime.timedelta()
        self.interface.register_result(tst_callable, uuid.uuid4().hex)
        self.interface.register_result(tst_add, uuid.uuid4().hex)
        entries = self.interface.count_results()
        assert entries == 2

    def test_delete_outdated_results(self):
        # register two results, one of them outdated.
        uuid_ = uuid.uuid4().hex
        self.interface.register_result(
            tst_callable,
            uuid_,
            status=sql_interface.TASK_STATUS_READY
        )
        # set ttl to 0:
        self.interface._result_ttl = datetime.timedelta()
        # this result is outdated
        self.interface.register_result(
            tst_add,
            uuid.uuid4().hex,
            status=sql_interface.TASK_STATUS_READY
        )
        entries = self.interface.count_results()
        assert entries == 2
        self.interface.delete_outdated_results()
        entries = self.interface.count_results()
        assert entries == 1
        # the remaining entry should be the `tst_callable` result
        entry = self.interface.get_result_by_uuid(uuid_)
        assert entry.function_module == tst_callable.__module__
        assert entry.function_name == tst_callable.__name__

    def test_delete_mixed_results(self):
        # register a waiting result, a regular result, an outdated result
        # and an outdated result with error state.
        # After deleting the outdated results the entries should be
        # decreased by one.
        # the waiting result:
        self.interface.register_result(tst_callable, uuid.uuid4().hex)
        # the regular result (not outdated)
        self.interface.register_result(
            tst_callable,
            uuid.uuid4().hex,
            status=sql_interface.TASK_STATUS_READY
        )
        # set ttl to 0:
        self.interface._result_ttl = datetime.timedelta()
        # the outdated result:
        self.interface.register_result(
            tst_callable,
            uuid.uuid4().hex,
            status=sql_interface.TASK_STATUS_READY
        )
        # the outdated result in error state:
        self.interface.register_result(
            tst_callable,
            uuid.uuid4().hex,
            status=sql_interface.TASK_STATUS_ERROR
        )
        # test for decreasing entries:
        entries = self.interface.count_results()
        self.interface.delete_outdated_results()
        remaining_entries = self.interface.count_results()
        assert entries - remaining_entries == 1

    def test_count_rows(self):
        # call _count_table_rows on an empty table:
        rows = self.interface._count_table_rows(
            table_name=sql_interface.DB_TABLE_NAME_TASK)
        assert rows == 0
        # create a row and count again:
        self.interface.register_callable(tst_callable)
        rows = self.interface._count_table_rows(
            table_name=sql_interface.DB_TABLE_NAME_TASK)
        assert rows == 1
        # check for exception in case of an unknown table:
        self.assertRaises(
            sqlite3.OperationalError,
            self.interface._count_table_rows,
            table_name="unknwon_table_name"
        )

    def test_count_results(self):
        # register three results.
        # check whether there are three entries in the database
        self.interface.register_result(tst_callable, uuid.uuid4().hex)
        self.interface.register_result(tst_add, uuid.uuid4().hex)
        self.interface.register_result(tst_multiply, uuid.uuid4().hex)
        entries = self.interface.count_results()
        assert entries == 3

    def test_count_tasks(self):
        # register three callables, two as cronjobs.
        # check whether there are three entries in the database
        self.interface.register_callable(tst_callable)
        self.interface.register_callable(tst_add, crontab="* * * * *")
        self.interface.register_callable(tst_multiply, crontab="* * * * *")
        entries = self.interface.count_tasks()
        assert entries == 3

    def test_delete_cronjobs(self):
        # register three callables, two as cronjobs.
        # delete the cronjobs and check that a single item in left
        # in the database.
        self.interface.register_callable(tst_callable)
        self.interface.register_callable(tst_add, crontab="* * * * *")
        self.interface.register_callable(tst_multiply, crontab="* * * * *")
        self.interface.delete_cronjobs()
        entries = self.interface.count_tasks()
        assert entries == 1
        # remaining entry should be the tst_callable
        entry = self.interface.get_tasks_on_due()[0]
        assert entry.function_module == tst_callable.__module__
        assert entry.function_name == tst_callable.__name__

    def test_initialize_settings_table(self):
        """
        Combined test for
            _initialize_settings_table()
            get_settings()
            set_settings()
        """

        def get_rows():
            return self.interface._count_table_rows(
                table_name=sql_interface.DB_TABLE_NAME_SETTINGS
            )
        # call to _initialize_settings_table() should add an entry:
        self.interface._initialize_settings_table()
        assert get_rows() == 1
        settings = self.interface.get_settings()
        assert settings.max_workers == sql_interface.DEFAULT_MAX_WORKERS
        new_max_workers = sql_interface.DEFAULT_MAX_WORKERS + 1
        settings.max_workers = new_max_workers
        self.interface.set_settings(settings)
        # value of max_workers should have changed
        new_settings = self.interface.get_settings()
        assert new_settings.max_workers == new_max_workers

        # a second call should have no effect:
        self.interface._initialize_settings_table()
        assert get_rows() == 1

        # and the former changed values are also unchanged:
        the_settings = self.interface.get_settings()
        assert the_settings.max_workers == new_max_workers

    def test_worker_settings(self):
        # increment and decrement worker_pids in the settings

        def check_settings():
            settings = self.interface.get_settings()
            assert settings.running_workers == len(test_pids)
            text = ",".join(map(str, test_pids))
            assert settings.worker_pids == text

        # register pids
        test_pids = [42, 377, 42980]
        for pid in test_pids:
            self.interface.increment_running_workers(pid)
        check_settings()
        # remove a single one
        pid = test_pids.pop(1)
        self.interface.decrement_running_workers(pid)
        check_settings()
        # remove the remaining ones
        while test_pids:
            self.interface.decrement_running_workers(test_pids.pop())
        check_settings()


# decorator testing includes database access.
# for easier testing decorator tests are included here.

class TestCronDecorator(unittest.TestCase):

    def setUp(self):
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(db_name=TEST_DB_NAME)
        self.orig_interface = decorators.interface
        decorators.interface = self.interface

    def tearDown(self):
        pathlib.Path(decorators.interface.db_name).unlink()
        decorators.interface = self.orig_interface

    def test_cron_no_arguments_active(self):
        # the database should have one entry with the default crontab
        wrapper = decorators.cron()
        func = wrapper(tst_cron)
        assert func == tst_cron
        entries = list(self.interface.get_tasks_by_signature(tst_cron))
        assert len(entries) == 1
        entry = entries[0]
        assert entry["crontab"] == decorators.DEFAULT_CRONTAB

    def test_suppress_identic_cronjobs(self):
        # register multiple cronjobs of a single callable.
        # then add the cronjob again by means of the decorator.
        # the db then should hold just a single entry deleting
        # the other ones.
        # should not happen:
        self.interface.register_callable(tst_cron, crontab=decorators.DEFAULT_CRONTAB)
        self.interface.register_callable(tst_cron, crontab=decorators.DEFAULT_CRONTAB)
        entries = list(self.interface.get_tasks_by_signature(tst_cron))
        assert len(entries) == 2
        # now add the same function with the cron decorator:
        crontab = "10 2 1 * *"
        wrapper = decorators.cron(crontab=crontab)
        func = wrapper(tst_cron)
        # just a single entry should now be in the database
        # (the one added by the decorator):
        entries = list(self.interface.get_tasks_by_signature(tst_cron))
        assert len(entries) == 1
        entry = entries[0]
        assert entry["crontab"] == crontab


class TestDelayDecorator(unittest.TestCase):

    def setUp(self):
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(db_name=TEST_DB_NAME)
        self.orig_decorator_interface = decorators.interface
        decorators.interface = self.interface

    def tearDown(self):
        pathlib.Path(self.interface.db_name).unlink()
        decorators.interface = self.orig_decorator_interface

    def _activate(self):
        self.interface.accept_registrations = True

    def _deactivate(self):
        self.interface.accept_registrations = False

    def test_wrapper_seen_by_worker(self):
        # does not return the original function but calls
        # the original function indirect instead of registering
        # the task in the db.
        self._deactivate()
        wrapper = decorators.delay(tst_delay)
        task_result = wrapper()
        assert task_result == 42
        # activate so that get_tasks_by_signature() works
        self._activate()
        entries = self.interface.get_tasks_by_signature(tst_delay)
        assert len(entries) == 0

    def test_active(self):
        wrapper = decorators.delay(tst_delay)
        wrapper_return_value = wrapper()
        assert isinstance(wrapper_return_value, TaskResult) is True
        entries = self.interface.get_tasks_by_signature(tst_delay)
        assert len(entries) == 1

    def test_inactive(self):
        # autocron does not run, but the wrapper returns a
        # TaskResult instance for the application.
        self.interface.autocron_lock_is_set = True
        wrapper = decorators.delay(tst_delay)
        wrapper_return_value = wrapper()
        assert isinstance(wrapper_return_value, TaskResult) is True
        assert wrapper_return_value.is_ready is True
        assert wrapper_return_value.result == 42

    def test_active_and_get_result(self):
        """
        Test story:

        1. wrap a function with the delay decorator.
           This should return a TaskResult.
        2. Check for the task entry in db.
        3. Then call `Worker.handle_tasks` what should return True.
        4. Then call `interface.get_result_by_uuid` which should return
           a TaskResult instance with the correct result.

        """
        # 1: wrap function and call the wrapper with arguments
        wrapper = decorators.delay(tst_add)
        task_result = wrapper(40, 2)
        assert task_result.uuid is not None
        assert isinstance(task_result.uuid, str)

        # 2: a single entry is now in both tables:
        time.sleep(0.001)  # have some patience with the db.
        task_entries = self.interface.get_tasks_by_signature(tst_add)
        assert len(task_entries) == 1
        # result is also of type TaskResult()
        result = self.interface.get_result_by_uuid(task_result.uuid)
        assert result is not None
        assert result.is_waiting is True

        # 3: let the worker handle the task:
        # instanciate but don't start the worker
        worker_ = worker.Worker(self.interface.db_name)
        # return True if at least one task has handled:
        return_value = worker_.handle_tasks()
        assert return_value is True
        time.sleep(0.001)  # have some patience with the db.
        # after handling the task should be removed from the db:
        task_entries = self.interface.get_tasks_by_signature(tst_add)
        assert len(task_entries) == 0

        # 4: check whether the worker has updated the result entry in the db:
        result = self.interface.get_result_by_uuid(task_result.uuid)
        assert result.is_ready is True
        assert result.result == 42  # 40 + 2


class TestHybridNamespace(unittest.TestCase):

    def setUp(self):
        self.data = {"pi": 3.141, "answer": 42}
        self.attr_dict = sql_interface.HybridNamespace(self.data)

    def test_dict_access(self):
        self.attr_dict["one"] = 1
        assert self.attr_dict["one"] == 1

    def test_attribute_access(self):
        self.attr_dict.two = 2
        assert self.attr_dict.two == 2

    def test_mixed_access(self):
        self.attr_dict.three = 3
        assert self.attr_dict["three"] == 3
        self.attr_dict["four"] = 4
        assert self.attr_dict.four == 4

    def test_get_init_values(self):
        assert self.attr_dict["pi"] == self.data["pi"]
        assert self.attr_dict.pi == self.data["pi"]
        assert self.attr_dict["answer"] == self.data["answer"]
        assert self.attr_dict.answer == self.data["answer"]


def task_result_function(a, b, c="c", d="d"):
    return "".join([a, b, c, d])


class TestTaskResult(unittest.TestCase):

    def setUp(self):
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(TEST_DB_NAME)

    def tearDown(self):
        if self.interface.db_name:
            pathlib.Path(self.interface.db_name).unlink()

    def test_task_result_from_function_call(self):
        args = ("a", "b")
        tr = sql_interface.TaskResult.from_function_call(
            task_result_function,
            *args
        )
        assert tr.status == sql_interface.TASK_STATUS_READY
        assert tr.result == "abcd"
        args = ("e", "f")
        kwargs = {"c": "g", "d": "h"}
        tr = sql_interface.TaskResult.from_function_call(
            task_result_function,
            *args,
            **kwargs
        )
        assert tr.result == "efgh"

    def test_update_task_result(self):
        """
        Create an empty TaskResult in waiting state and create a result.
        Both with the same uuid. Check whether TaskResult can update
        itself with a delated result.
        """
        uid = uuid.uuid4().hex
        tr = sql_interface.TaskResult.from_registration(uid, self.interface)
        assert tr.status is sql_interface.TASK_STATUS_WAITING
        assert tr.is_waiting is True
        # next is a hack because self.assertRaises expects as second argument
        # a callable, like self.assertRaises(AttributeError, tr.result).
        # This does not work with properties.
        try:
            tr.result
        except AttributeError:
            exception_was_raised = True
        else:
            exception_was_raised = False
        assert exception_was_raised is True
        # now inject the task to execute:
        self.interface.register_result(task_result_function, uid, ("a", "b"))
        # the task is still in waiting state:
        assert tr.is_waiting is True
        # update result, we calculate the result here:
        result = task_result_function("a", "b")
        self.interface.update_result(uid, result=result)
        # now the state hase changed:
        assert tr.is_waiting is False
        assert tr.is_ready is True
        assert tr.result == "abcd"


class TestDelayedInitialization(unittest.TestCase):
    """
    Special case of the TestCronDecorator suite assuming that the name
    of the database will get provided by the start() function which will
    get called after the @con-decorators have been executed at import
    time. So these tests are testing the preregistration of callables.
    """

    def setUp(self):
        # set up a non-initialized instance of the sql_interface,
        self.orig_interface = decorators.interface
        decorators.interface = sql_interface.SQLiteInterface()

    def tearDown(self):
        if decorators.interface.db_name is not None:
            pathlib.Path(decorators.interface.db_name).unlink()
        decorators.interface = self.orig_interface

    def test_preregistration(self):
        # register a cronfunction (defined near TestCronDecorator()) before
        # the database has been set up.
        decorators.interface.register_callable(tst_cron)
        # should fail, because the database has not been set up yet:
        self.assertRaises(OSError, decorators.interface.get_tasks)
        # after initializing the task should be available:
        decorators.interface.init_database(TEST_DB_NAME)
        tasks = decorators.interface.get_tasks()
        assert len(tasks) == 1
        # check that the task is indeed the tst_cron()
        # (the tasks are of type 'HybridNamespace'. See worker.process_task)
        task = tasks[0]
        assert task.function_name == tst_cron.__name__

    def test_preregister_tst_cron(self):
        # use the cron decorator to register cron functions
        cron_decorator = decorators.cron()
        cron_decorator(tst_cron)
        cron_decorator(tst_delay)
        # set the database and check for two entries
        decorators.interface.init_database(TEST_DB_NAME)
        decorators.interface._register_preregistered_tasks()
        tasks = decorators.interface.get_tasks()
        assert len(tasks) == 2
        # register the same tst_cron() again
        # should not duplicate the cron-tasks:
        cron_decorator(tst_cron)
        tasks = decorators.interface.get_tasks()
        assert len(tasks) == 2
        # registration of other functions should work as expected:
        decorators.interface.register_callable(tst_multiply)
        tasks = decorators.interface.get_tasks()
        assert len(tasks) == 3

    def test_register_tst_cron_after_start(self):
        decorators.interface.init_database(TEST_DB_NAME)
        tasks = decorators.interface.get_tasks()
        assert len(tasks) == 0
        cron_decorator = decorators.cron()
        cron_decorator(tst_cron)
        tasks = decorators.interface.get_tasks()
        assert len(tasks) == 1
        # is the crontab attribute set?
        task = tasks[0]
        self.assertTrue(task.crontab)




"""
sql_interface.py

SQLite interface for storing tasks.

The autocron database consists of three tables:
task: stores all tasks that should get executed later
result: stores all results from task returning a result
settings: configuration settings for a project
"""

import datetime
import pathlib
import pickle
import sqlite3
import types


DEFAULT_STORAGE = ".autocron"


# -- table: task - structure and commands ------------------------------------
DB_TABLE_NAME_TASK = "task"
CMD_CREATE_TASK_TABLE = f"""
CREATE TABLE IF NOT EXISTS {DB_TABLE_NAME_TASK}
(
    uuid TEXT,
    schedule datetime,
    status INTEGER,
    crontab TEXT,
    function_module TEXT,
    function_name TEXT,
    function_arguments BLOB
)
"""
CMD_STORE_TASK = f"""
INSERT INTO {DB_TABLE_NAME_TASK} VALUES
(
    :uuid,
    :schedule,
    :status,
    :crontab,
    :function_module,
    :function_name,
    :function_arguments
)
"""
TASK_COLUMN_SEQUENCE =\
    "rowid,uuid,schedule,status,crontab,"\
    "function_module,function_name,function_arguments"
CMD_GET_TASKS_BY_NAME =f"""
    SELECT {TASK_COLUMN_SEQUENCE} FROM {DB_TABLE_NAME_TASK}
    WHERE function_module == ? AND function_name == ?"""
CMD_GET_TASKS_ON_DUE = f"""
    SELECT {TASK_COLUMN_SEQUENCE} FROM {DB_TABLE_NAME_TASK}
    WHERE schedule <= ?"""
CMD_GET_TASKS_ON_DUE_WITH_STATUS = CMD_GET_TASKS_ON_DUE + " AND status == ?"
CMD_GET_TASKS = f"""
    SELECT {TASK_COLUMN_SEQUENCE} FROM {DB_TABLE_NAME_TASK}"""
CMD_UPDATE_TASK_STATUS = f"""
    UPDATE {DB_TABLE_NAME_TASK} SET status = ? WHERE rowid == ?"""
CMD_GET_CRONTASKS = f"""
    SELECT  {TASK_COLUMN_SEQUENCE} FROM {DB_TABLE_NAME_TASK}
    WHERE crontab <> ''"""
CMD_UPDATE_CRONTASK_SCHEDULE = f"""
    UPDATE {DB_TABLE_NAME_TASK} SET schedule = ?, status = ? WHERE rowid == ?"""
CMD_DELETE_TASK = f"DELETE FROM {DB_TABLE_NAME_TASK} WHERE rowid == ?"
CMD_DELETE_CRON_TASKS = f"DELETE FROM {DB_TABLE_NAME_TASK} WHERE crontab <> ''"
CMD_COUNT_TABLE_ROWS = "SELECT COUNT(*) FROM {table_name}"

# Status codes used for task-status the result-entries:
TASK_STATUS_WAITING = 1
TASK_STATUS_PROCESSING = 2
TASK_STATUS_READY = 3
TASK_STATUS_ERROR = 4
TASK_STATUS_UNAVAILABLE = 5


# -- table: result - structure and commands ----------------------------------
DB_TABLE_NAME_RESULT = "result"
CMD_CREATE_RESULT_TABLE = f"""
CREATE TABLE IF NOT EXISTS {DB_TABLE_NAME_RESULT}
(
    uuid TEXT PRIMARY KEY,
    status INTEGER,
    function_module TEXT,
    function_name TEXT,
    function_arguments BLOB,
    function_result BLOB,
    error_message TEXT,
    ttl datetime
)
"""

# Storage time (time to live) for results in seconds
RESULT_TTL = 1800

CMD_STORE_RESULT = f"""
INSERT INTO {DB_TABLE_NAME_RESULT} VALUES
(
    :uuid,
    :status,
    :function_module,
    :function_name,
    :function_arguments,
    :function_result,
    :error_message,
    :ttl
)
"""
RESULT_COLUMN_SEQUENCE =\
    "rowid,uuid,status,function_module,function_name,"\
    "function_arguments,function_result,error_message, ttl"
CMD_GET_RESULTS = f"""
    SELECT {RESULT_COLUMN_SEQUENCE} FROM {DB_TABLE_NAME_RESULT}"""
CMD_GET_RESULT_BY_UUID = f"""
    SELECT {RESULT_COLUMN_SEQUENCE} FROM {DB_TABLE_NAME_RESULT}
    WHERE uuid == ?"""
CMD_UPDATE_RESULT = f"""
    UPDATE {DB_TABLE_NAME_RESULT} SET
        status = ?,
        function_result = ?,
        error_message = ?,
        ttl = ?
    WHERE uuid == ?"""
# CMD_DELETE_RESULT = f"""\
#     DELETE FROM {DB_TABLE_NAME_RESULT} WHERE uuid == ?"""
CMD_DELETE_OUTDATED_RESULTS = f"""
    DELETE FROM {DB_TABLE_NAME_RESULT}
    WHERE status == {TASK_STATUS_READY} AND ttl <= ?"""


# -- table: settings - structure and commands --------------------------------
DB_TABLE_NAME_SETTINGS = "settings"
CMD_CREATE_SETTINGS_TABLE = f"""
CREATE TABLE IF NOT EXISTS {DB_TABLE_NAME_SETTINGS}
(
    max_workers INTEGER,
    running_workers INTEGER,
    monitor_lock INTEGER,
    autocron_lock INTEGER,
    monitor_idle_time REAL,
    worker_idle_time REAL,
    worker_pids TEXT,
    result_ttl INTEGER
)
"""

DEFAULT_MAX_WORKERS = 1
DEFAULT_RUNNING_WORKERS = 0
DEFAULT_MONITOR_LOCK = 0
DEFAULT_AUTOCRON_LOCK = 0
DEFAULT_MONITOR_IDLE_TIME = 5.0  # seconds
DEFAULT_WORKER_IDLE_TIME = 2.0  # seconds
DEFAULT_WORKER_PIDS = ""

CMD_SETTINGS_STORE_VALUES = f"""
INSERT INTO {DB_TABLE_NAME_SETTINGS} VALUES
(
    :max_workers,
    :running_workers,
    :monitor_lock,
    :autocron_lock,
    :monitor_idle_time,
    :worker_idle_time,
    :worker_pids,
    :result_ttl
)
"""
SETTINGS_COLUMN_SEQUENCE =\
    "rowid,max_workers,running_workers,"\
    "monitor_lock,autocron_lock,"\
    "monitor_idle_time,worker_idle_time,"\
    "worker_pids,result_ttl"
BOOLEAN_SETTINGS = ["monitor_lock", "autocron_lock"]
CMD_SETTINGS_GET_SETTINGS = f"""
    SELECT {SETTINGS_COLUMN_SEQUENCE} FROM {DB_TABLE_NAME_SETTINGS}"""
CMD_SETTINGS_UPDATE = f"""
    UPDATE {DB_TABLE_NAME_SETTINGS} SET
        max_workers = ?,
        running_workers = ?,
        monitor_lock = ?,
        autocron_lock = ?,
        monitor_idle_time = ?,
        worker_idle_time = ?,
        worker_pids = ?,
        result_ttl = ?
    WHERE rowid == ?"""


# sqlite3 default adapters and converters deprecated as of Python 3.12:

def datetime_adapter(value):
    """
    Gets a python datetime-instance and returns an ISO 8601 formated
    string for sqlite3 storage.
    """
    return value.isoformat()


def datetime_converter(value):
    """
    Gets an ISO 8601 formated byte-string (from sqlite3) and returns a
    python datetime datatype.
    """
    return datetime.datetime.fromisoformat(value.decode())


sqlite3.register_adapter(datetime.datetime, datetime_adapter)
sqlite3.register_converter("datetime", datetime_converter)


# pylint does not like instances with dynamic attributes:
# pylint: disable=attribute-defined-outside-init
class HybridNamespace(types.SimpleNamespace):
    """
    A namespace-object with additional dictionary-like attribute access.
    """
    def __init__(self, data=None):
        """
        Set initial values.
        If data is given, it must be a dictionary.
        """
        if data is not None:
            self.__dict__.update(data)

    def __getitem__(self, name):
        return self.__dict__[name]

    def __setitem__(self, name, value):
        self.__dict__[name] = value

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        return "\n".join(f"{k}:{v}" for k, v in self.__dict__.items())

    def __str__(self):
        # same as __repr__ but without the rowid
        lines = repr(self).split("\n")
        lines = [line for line in lines if not line.startswith("rowid")]
        return "\n".join(lines)


class TaskResult(HybridNamespace):
    """
    Helper class to make task-results more handy.
    """

    def _refresh(self):
        """
        If the status is TASK_STATUS_WAITING try to update the
        task_result dictionary with data retrieved from the database.
        This will return a new task_result instance. If the new instance
        is still in waiting status, do nothing. Otherwise update
        self.__dict__ with the retrieved data.
        The `interface` argument is for testing.
        """
        if self.status == TASK_STATUS_WAITING:
            try:
                result = self.interface.get_result_by_uuid(self.uuid)
            except AttributeError:
                pass
            else:
                if result is not None:
                    self.__dict__.update(result.__dict__)

    @property
    def result(self):
        """
        Shortcut to access the result. If the result is not available
        because the task still waits to get executed an AttributeError
        with the message "result not available" is raised.
        """
        if self.status == TASK_STATUS_READY:
            return self.function_result
        raise AttributeError("result not available.")

    @property
    def is_waiting(self):
        """indicates task still waiting to get processed."""
        self._refresh()
        return self.status == TASK_STATUS_WAITING

    @property
    def is_ready(self):
        """indicates task has been processed."""
        self._refresh()
        return self.status == TASK_STATUS_READY

    @property
    def has_error(self):
        """indicates error_message is set."""
        self._refresh()
        return self.status == TASK_STATUS_ERROR

    @classmethod
    def from_data_tuple(cls, row_data):
        """
        Returns a new TaskResult-Instance initialized with a tuple
        representing the data from result-table row.
        """
        column_names = RESULT_COLUMN_SEQUENCE.split(",")
        data = dict(zip(column_names, row_data))
        instance = cls(data)
        instance.function_result = pickle.loads(instance.function_result)
        instance.function_arguments = pickle.loads(instance.function_arguments)
        return instance

    @classmethod
    def from_function_call(cls, func, *args, **kwargs):
        """
        Renturns a new TaskResult-Instance with the result from the
        given function executed with the given arguments. This exists
        for type consistency to return a TaskResult from delay-decorated
        functions even if autotask is inactive.
        """
        data = {
            "function_result": func(*args, **kwargs),
            "function_arguments": (args, kwargs),
            "status": TASK_STATUS_READY
        }
        return cls(data)

    @classmethod
    def from_registration(cls, uuid, interface):
        """
        Returns a TaskResult-Instance with the state
        TASK_STATUS_WAITING, the result None and a set uuid. This
        instance can be used to call the ``update()`` method for a
        delayed result.
        """
        data = {
            "function_result": None,
            "status": TASK_STATUS_WAITING,
            "uuid": uuid,
            "interface": interface
        }
        return cls(data)

    @classmethod
    def from_none(cls):
        """
        Returns a new empty TaskResult-Instance if there is no waiting
        function to call and no result to expect. The available
        attributes are 'status' with the value TASK_STATUS_UNAVAILABLE and
        the 'result' has the value None. All other attributes are
        undefined and accessing them will raise an AttributeError.
        """
        data = {
            "status": TASK_STATUS_UNAVAILABLE,
            "function_result": None
        }
        return cls(data)


# pylint: disable=too-many-public-methods
class SQLiteInterface:
    """
    SQLite interface for application specific operations.
    This class is a Singleton.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        self._preregistered_tasks = []
        self._result_ttl = datetime.timedelta(seconds=RESULT_TTL)
        self._accept_registrations = True
        self._db_name = None
        self.autocron_lock_is_set = None

    @property
    def db_name(self):
        """Return hidden _db_name (because of the setter)"""
        return self._db_name

    @db_name.setter
    def db_name(self, db_name):
        """
        Expand the given database filename to the storage-location path
        if the filename is not an absolute path. In the latter case the
        absolute path will get used (and must exist). If db_name is None
        set None directly.
        """
        if db_name is None:
            self._db_name = None
        else:
            self._set_storage_location(db_name)

    @property
    def accept_registrations(self):
        """
        Return a boolean whether callables are allowed to get registered
        in the database.
        """
        return self._accept_registrations and not self.autocron_lock_is_set

    @accept_registrations.setter
    def accept_registrations(self, value):
        """
        Setter for the hidden _accept_registrations flag (because of the
        getter)
        """
        self._accept_registrations = value

    @property
    def is_initialized(self):
        """Flag whether the database is available."""
        return self._db_name is not None

    def _init_database(self):
        """
        Creates a new database in case that an older one is not existing
        and in case of missing settings set the settings-default values.
        """
        self._create_tables()
        self._initialize_settings_table()
        settings = self.get_settings()
        self.autocron_lock_is_set = settings.autocron_lock
        self._result_ttl = datetime.timedelta(seconds=settings.result_ttl)

    def _create_tables(self):
        """
        Create all used tables in case of a new db and missing tables.
        """
        self._execute(CMD_CREATE_TASK_TABLE)
        self._execute(CMD_CREATE_RESULT_TABLE)
        self._execute(CMD_CREATE_SETTINGS_TABLE)

    def _initialize_settings_table(self):
        """
        Check for an existing settings row in the settings-table.
        If there is no row create an entry with the default values.
        """
        rows = self._count_table_rows(DB_TABLE_NAME_SETTINGS)
        if not rows:
            data = {
                "max_workers": DEFAULT_MAX_WORKERS,
                "running_workers": DEFAULT_RUNNING_WORKERS,
                "monitor_lock": DEFAULT_MONITOR_LOCK,
                "autocron_lock": DEFAULT_AUTOCRON_LOCK,
                "monitor_idle_time": DEFAULT_MONITOR_IDLE_TIME,
                "worker_idle_time": DEFAULT_WORKER_IDLE_TIME,
                "worker_pids": DEFAULT_WORKER_PIDS,
                "result_ttl": RESULT_TTL
            }
            self._execute(CMD_SETTINGS_STORE_VALUES, data)

    def _count_table_rows(self, table_name):
        """
        Helper function to count the number of entries in the given
        table. Returns a numeric value. In case of an unknown table_name
        a sqlite3.OperationalError will get raised.
        """
        cmd = CMD_COUNT_TABLE_ROWS.format(table_name=table_name)
        cursor = self._execute(cmd)
        number_of_rows = cursor.fetchone()[0]
        return number_of_rows

    def _execute(self, cmd, parameters=(), many=False):
        """
        Run a command with parameters. Parameters can be a sequence of
        values to get used in an ordered way or a dictionary with
        key-value pairs, where the key are the value-names used in the
        db (i.e. the column names).
        If 'many' is true then con.executemany() gets called and
        parameters is interpreted differently as as sequence of ordered
        tuples or dictionaries as placehoöders for the provided cmd.
        """
        if self._db_name is None:
            raise IOError("No autocron database defined.")
        con = sqlite3.connect(
            self._db_name,
            detect_types=sqlite3.PARSE_DECLTYPES
        )
        with con:
            if many:
                return con.executemany(cmd, parameters)
            return con.execute(cmd, parameters)

    def _set_storage_location(self, db_name):
        """
        Set the database file location. If the path is absolute use the
        path as is. The directory part of the path must exist. If the
        path is relative the database file gets stored in the
        '~.autocron/' directory. If no home directory is found on the
        running platform the current working directory is used as a
        fallback. However in such cases it would be safer to provide an
        absolute path to the database location.
        """
        path = pathlib.Path(db_name)
        if not path.is_absolute():
            try:
                path = pathlib.Path.home() / DEFAULT_STORAGE / path.name
            except RuntimeError:
                # no home directory found
                path = pathlib.Path.cwd() / db_name
        self._db_name = path

    def _preregister_task(self, data):
        """
        Take the data, which is a dictionary,and convert it to another
        dict without the key 'self'. Pickle the dict and append it to
        _preregistered tasks.
        """
        data = {k: v for k, v in data.items() if k != "self"}
        self._preregistered_tasks.append(data)

    def _register_preregistered_tasks(self):
        """
        Run the stored registrations on the now up and
        running database.
        """
        for data in self._preregistered_tasks:
            self.register_callable(**data)

    def init_database(self, db_name):
        """
        Public callable for delayed initialization. Set the database
        (which is a string or Path) and runs the correponding
        initialization.
        """
        if not self.is_initialized:
            self._set_storage_location(db_name)
            self._init_database()
            # ignore the result, but set new state:
            self.get_tasks_on_due(
                status=TASK_STATUS_PROCESSING,
                new_status=TASK_STATUS_WAITING
            )
            self._register_preregistered_tasks()


    # -- task-methods ---

    @staticmethod
    def _fetch_all_callable_entries(cursor):
        """
        Internal function to iterate over a selection of entries and unpack
        the columns to a dictionary with the following key-value pairs:

            {
                "rowid": integer,
                "uuid": string,
                "schedule": datetime,
                "status": integer,
                "crontab": string,
                "function_module": string,
                "function_name": string,
                "args": tuple(of original datatypes),
                "kwargs": dict(of original datatypes),
            }

        Returns a list of HybridNamespace instances or an empty list if
        a selection does not match any row.
        """
        def process(row):
            """
            Gets a `row` and returns a dictionary with Python datatypes.
            `row` is an ordered tuple of columns as defined in `CREATE
            TABLE`. The blob column with the pickled arguments is the
            last column.
            """
            args, kwargs = pickle.loads(row[-1])
            data = {
                key: row[i] for i, key in enumerate(
                    TASK_COLUMN_SEQUENCE.strip().split(",")[:-1]
                )
            }
            data["args"] = args
            data["kwargs"] = kwargs
            return HybridNamespace(data)

        return [process(row) for row in cursor.fetchall()]

    # pylint: disable=too-many-arguments
    def register_callable(
        self,
        func,
        uuid="",
        schedule=None,
        status=TASK_STATUS_WAITING,
        crontab="",
        args=None,
        kwargs=None,
        unique=False,
    ):
        """
        Store a callable in the task-table of the database. If the
        argument `unique` is given, don't register a callable twice. In
        this case an already registered callable with the same
        signature (module.name) gets overwritten. This can be useful for
        cron-tasks to not register them multiple times.
        """
        if self.is_initialized:
            if unique:
                tasks = self.get_tasks_by_signature(func)
                for task in tasks:
                    self.delete_callable(task)
            if not schedule:
                schedule = datetime.datetime.now()
            if args is None:
                args = ()
            if kwargs is None:
                kwargs = {}
            arguments = pickle.dumps((args, kwargs))
            data = {
                "uuid": uuid,
                "schedule": schedule,
                "status": status,
                "crontab": crontab,
                "function_module": func.__module__,
                "function_name": func.__name__,
                "function_arguments": arguments,
            }
            self._execute(CMD_STORE_TASK, data)
            if uuid:
                self.register_result(func, uuid, args=args, kwargs=kwargs)
        else:
            self._preregister_task(locals())

    def get_tasks(self):
        """
        Generic method to return all tasks as a list of HybridNamespace
        instances.
        """
        cursor = self._execute(CMD_GET_TASKS)
        return self._fetch_all_callable_entries(cursor)

    def get_tasks_on_due(self, schedule=None, status=None, new_status=None):
        """
        Returns tasks on due as a list of HybridNamespace instances. If
        'status' is given the status is also used for the selection. If
        'new_status' is given the selection gets updated with the new
        status.
        """
        if not schedule:
            schedule = datetime.datetime.now()
        if status:
            cursor = self._execute(
                CMD_GET_TASKS_ON_DUE_WITH_STATUS,
                [schedule, status]
            )
        else:
            cursor = self._execute(CMD_GET_TASKS_ON_DUE, [schedule])
        tasks = self._fetch_all_callable_entries(cursor)
        if new_status and tasks:
            values = [(new_status, task.rowid) for task in tasks]
            self._execute(CMD_UPDATE_TASK_STATUS, values, many=True)
            # also update the status in the previous fetched tasks:
            for task in tasks:
                task.status = new_status
        return tasks

    def get_tasks_by_signature(self, func):
        """
        Return all tasks matching the function-signature as a list of
        HybridNamespace instances.
        """
        parameters = func.__module__, func.__name__
        cursor = self._execute(CMD_GET_TASKS_BY_NAME, parameters)
        return self._fetch_all_callable_entries(cursor)

    def delete_callable(self, entry):
        """
        Delete the entry in the callable-table. Entry should be a
        dictionary as returned from `get_tasks_on_due()`. The row to delete
        gets identified by the `rowid`.
        """
        self._execute(CMD_DELETE_TASK, [entry["rowid"]])

    def get_crontasks(self):
        """
        Return all crontasks as a list of HybridNamespace instances.
        """
        cursor = self._execute(CMD_GET_CRONTASKS)
        return self._fetch_all_callable_entries(cursor)

    def delete_cronjobs(self):
        """
        Delete all cronjobs from the task-table.
        """
        self._execute(CMD_DELETE_CRON_TASKS)

    def update_crontask_schedule(self, rowid, schedule):
        """
        Update the `schedule` of the table entry with the given `rowid`.
        As this should be a crontask the status is set to WAITING.
        """
        parameters = schedule, TASK_STATUS_WAITING, rowid
        self._execute(CMD_UPDATE_CRONTASK_SCHEDULE, parameters)

    def count_tasks(self):
        """
        Returns the number of rows in the task-table, therefore
        providing the number of tasks stored in the database.
        """
        return self._count_table_rows(DB_TABLE_NAME_TASK)


    # -- result-methods ---

    @property
    def result_ttl(self):
        """
        Returns the new ttl as a datetime instance with an offset of now.
        """
        return datetime.datetime.now() + self._result_ttl

    def register_result(
            self,
            func,
            uuid,
            args=(),
            status=TASK_STATUS_WAITING,
            kwargs=None,
        ):
        """
        Register an entry in the result table of the database. The entry
        stores the uuid and the status `False` as zero `0` because the
        task is pending and no result available jet.
        """
        if not kwargs:
            kwargs = {}
        arguments = pickle.dumps((args, kwargs))
        data = {
            "uuid": uuid,
            "status": status,
            "function_module": func.__module__,
            "function_name": func.__name__,
            "function_arguments": arguments,
            "function_result": pickle.dumps(None),
            "error_message": "",
            "ttl": self.result_ttl,
        }
        self._execute(CMD_STORE_RESULT, data)

    def get_results(self):
        """
        Generic method to return all results as a list of TaskResult
        instances.
        """
        cursor = self._execute(CMD_GET_RESULTS)
        rows = cursor.fetchall()
        return [TaskResult.from_data_tuple(row) for row in rows]

    def get_result_by_uuid(self, uuid):
        """
        Return a dataset (as TaskResult) or None.
        """
        cursor = self._execute(CMD_GET_RESULT_BY_UUID, (uuid,))
        row = cursor.fetchone()  # tuple of data or None
        if row:
            result = TaskResult.from_data_tuple(row)
        else:
            result = None
        return result

    def update_result(self, uuid, result=None, error_message=""):
        """
        Updates the result-entry with the given `uuid` to status 1|2 and
        stores the `result` or `error_message`.
        """
        status = TASK_STATUS_ERROR if error_message else TASK_STATUS_READY
        function_result = pickle.dumps(result)
        ttl = self.result_ttl
        parameters = status, function_result, error_message, ttl, uuid
        self._execute(CMD_UPDATE_RESULT, parameters)

    def count_results(self):
        """
        Returns the number of rows in the task-table, therefore
        providing the number of tasks stored in the database.
        """
        return self._count_table_rows(DB_TABLE_NAME_RESULT)

    def delete_outdated_results(self):
        """
        Deletes results with status TASK_STATUS_READY that have exceeded
        the time to live (ttl).
        """
        now = datetime.datetime.now()
        self._execute(CMD_DELETE_OUTDATED_RESULTS, (now,))

    # -- setting-methods ---

    def get_settings(self):
        """
        Returns a HybridNamespace instance with the settings as attributes:
        - max_workers
        - running_workers
        - monitor_lock
        - autocron_lock
        - monitor_idle_time
        - worker_idle_time
        - worker_pids
        - result_ttl
        - rowid (not a setting but included)
        """
        cursor = self._execute(CMD_SETTINGS_GET_SETTINGS)
        row = cursor.fetchone()  # there is only one row
        col_names = SETTINGS_COLUMN_SEQUENCE.split(",")
        data = dict(zip(col_names, row))
        for key in BOOLEAN_SETTINGS:
            data[key] = bool(data[key])
        return HybridNamespace(data)

    def set_settings(self, settings):
        """
        Takes a HybridNamespace instance as settings
        argument (like the one returned from get_settings) and updates
        the setting values in the database.
        """
        data = (
            settings.max_workers,
            settings.running_workers,
            int(settings.monitor_lock),
            int(settings.autocron_lock),
            settings.monitor_idle_time,
            settings.worker_idle_time,
            settings.worker_pids,
            settings.result_ttl,
            settings.rowid
        )
        self._execute(CMD_SETTINGS_UPDATE, data)

    def get_monitor_idle_time(self):
        """
        Convenience function to get the monitor_idle_time from the settings.
        """
        return self.get_settings().monitor_idle_time

    def get_worker_idle_time(self):
        """
        Convenience function to get the worker_idle_time from the settings.
        """
        return self.get_settings().worker_idle_time

    def get_max_workers(self):
        """
        Convenience function to get the max_workers from the settings.
        """
        return self.get_settings().max_workers

    @property
    def monitor_lock_flag_is_set(self):
        """
        Returns the status of the monitor-lock flag..
        """
        return self.get_settings().monitor_lock

    def set_monitor_lock_flag(self, value):
        """
        Set monitor_lock flag to the given state.
        """
        settings = self.get_settings()
        settings.monitor_lock = value
        self.set_settings(settings)

    def increment_running_workers(self, pid):
        """
        Increment the running_worker-setting by 1.
        """
        settings = self.get_settings()
        if settings.worker_pids:
            pids = settings.worker_pids.split(",")
        else:
            pids = []
        pids.append(str(pid))
        settings.worker_pids = ",".join(pids)
        settings.running_workers += 1
        self.set_settings(settings)

    def decrement_running_workers(self, pid):
        """
        Decrement the running_worker-setting by 1.
        But don't allow a value below zero.
        """
        settings = self.get_settings()
        if settings.worker_pids:
            pids = settings.worker_pids.split(",")
        else:
            pids = []
        try:
            pids.remove(str(pid))
        except ValueError:
            # can happen when decrement gets called before increment
            # otherwise it is a weird error that should not happen
            pass
        else:
            if settings.running_workers > 0:
                settings.running_workers -= 1
            settings.worker_pids = ",".join(pids)
            self.set_settings(settings)



"""
test_engine.py

tests for the engine and the worker.
"""

import pathlib
import subprocess
import sys
import threading
import time
import warnings
import unittest

from autocron import engine
from autocron import sql_interface
from autocron import worker


TEST_DB_NAME = "test.db"


# all labels starting with 'tst_' are test-functions:

def tst_cron():
    pass


class TestEngine(unittest.TestCase):

    def setUp(self):
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.engine = engine.Engine(interface=self.interface)

    def tearDown(self):
        # clean up if tests don't run through
        if self.interface.db_name:
            pathlib.Path(self.interface.db_name).unlink()

    def test_start_subprocess(self):
        process = engine.start_subprocess(self.interface.db_name)
        assert isinstance(process, subprocess.Popen) is True
        assert process.poll() is None
        process.terminate()
        time.sleep(0.02)  # give process some time to terminate
        assert process.poll() is not None

    def test_start_is_allowed(self):
        # don't start a thread here:
        self.engine.monitor_thread = "some reference"
        result = self.engine.start(TEST_DB_NAME)
        # on default the  autocron- and monitor-lock flags are False:
        assert result is True
        # starting a second time should not work because the
        # monitor-lock flag should be set now:
        result = self.engine.start(TEST_DB_NAME)
        assert result is False
        # check direct for the monitor_lock flag:
        self.assertTrue(self.engine.interface.monitor_lock_flag_is_set)
        # check for releasing the flag on stop
        self.engine.stop()
        self.assertFalse(self.engine.interface.monitor_lock_flag_is_set)


class TestAutocronFlagInjection(unittest.TestCase):
    """
    Test that the autocron_lock flag from the settings is readable by
    the SQLiteInterface. Also test whether SQLiteInterface indeed
    behaves like a singleton.
    """

    def test_new_interface_instance(self):
        sql_interface.SQLiteInterface._instance = None
        interface1 = sql_interface.SQLiteInterface()
        interface2 = sql_interface.SQLiteInterface()
        assert interface1 is interface2
        sql_interface.SQLiteInterface._instance = None
        interface3 = sql_interface.SQLiteInterface()
        assert interface1 is not interface3

    def test_injection(self):
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(db_name=TEST_DB_NAME)
        assert self.interface.autocron_lock_is_set is False
        settings = self.interface.get_settings()
        settings.autocron_lock = True
        self.interface.set_settings(settings)
        # get new instance with existing db
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(db_name=TEST_DB_NAME)
        assert self.interface.autocron_lock_is_set is True
        pathlib.Path(self.interface.db_name).unlink()


class TestAutocronFlag(unittest.TestCase):
    """
    Injects the autocron_lock flag set to True in the database: the
    engine should not start.
    """

    def setUp(self):
        # create self.interface twice to inject the autocron-flag:
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(db_name=TEST_DB_NAME)
        settings = self.interface.get_settings()
        settings.autocron_lock = True
        self.interface.set_settings(settings)
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.engine = engine.Engine(interface=self.interface)

    def tearDown(self):
        # clean up if tests don't run through
        pathlib.Path(self.interface.db_name).unlink()

    def test_autocron_flag(self):
        self.engine.monitor_thread = "some reference"
        result = self.engine.start(TEST_DB_NAME)
        assert result is False


class TestWorkerStartStop(unittest.TestCase):

    def setUp(self):
        # set worker_idle_time to a low value
        self._worker_idle_time = 0.01
        self._monitor_idle_time = 0.01
        sql_interface.SQLiteInterface._instance = None
        self.interface = sql_interface.SQLiteInterface()
        self.interface.init_database(db_name=TEST_DB_NAME)
        settings = self.interface.get_settings()
        settings.worker_idle_time = self._worker_idle_time
        settings.worker_idle_time = self._monitor_idle_time
        self.interface.set_settings(settings)
        self.cmd = [sys.executable, worker.__file__, self.interface.db_name]
        self.cwd = pathlib.Path.cwd()
        warnings.simplefilter("ignore", ResourceWarning)

    def tearDown(self):
        # clean up if tests don't run through
        pathlib.Path(self.interface.db_name).unlink(missing_ok=True)

    def test_start_and_stop_workerprocess(self):
        process = subprocess.Popen(self.cmd, cwd=self.cwd)
        assert process.poll() is None  # subprocess runs
        process.terminate()
        time.sleep(self._worker_idle_time * 2)
        assert process.poll() is not None


    def test_delete_crontasks_on_shutdown(self):
        """
        Scenario: put a cron-task in the database. Then close the
        database (to simulate a shutdown of the application). The
        cron-task should be deleted.
        """
        # put a crontask in the database:
        self.interface.register_callable(tst_cron, crontab="* * * * *")
        tasks = self.interface.get_tasks_by_signature(tst_cron)
        self.assertTrue(tasks)

        # simulate an engine in running mode and stop it:
        engine_ = engine.Engine(interface=self.interface)
        engine_.monitor_thread = True
        engine_.stop()

        # the crontask should now be deleted:
        tasks = self.interface.get_tasks_by_signature(tst_cron)
        self.assertFalse(tasks)
